import datetime

import ulid
from sqlmodel import Field, SQLModel


class MyEntityBase(SQLModel):
    my_column: str


class MyEntityCreate(MyEntityBase):
    pass


class MyEntityUpdate(SQLModel):
    my_column: str | None = None


class MyEntityResponse(MyEntityBase):
    id: str


class MyEntity(SQLModel, table=True):
    id: str = Field(default_factory=lambda: ulid.new().str, primary_key=True)
    created_at: datetime.datetime
    created_by: str
    updated_at: datetime.datetime
    updated_by: str
    my_column: str



import os

from fastapp_template._zrb.config import APP_DIR
from fastapp_template._zrb.group import app_create_group
from fastapp_template._zrb.helper import (
    get_existing_module_names,
    get_existing_schema_names,
)
from fastapp_template._zrb.input import (
    existing_module_input,
    new_entity_column_input,
    new_entity_input,
    plural_entity_input,
)

from zrb import AnyContext, Scaffolder, Task, make_task
from zrb.util.codemod.add_code_to_module import add_code_to_module
from zrb.util.codemod.add_parent_to_class import add_parent_to_class
from zrb.util.string.conversion import to_pascal_case, to_snake_case


@make_task(
    name="validate-create-my-app-name-entity",
    input=[
        existing_module_input,
        new_entity_input,
        plural_entity_input,
        new_entity_column_input,
    ],
    retries=0,
)
async def validate_create_my_app_name_entity(ctx: AnyContext):
    module_name = to_snake_case(ctx.input.module)
    if module_name not in get_existing_module_names():
        raise ValueError(f"Module not exist: {module_name}")
    schema_name = to_snake_case(ctx.input.entity)
    if schema_name in get_existing_schema_names():
        raise ValueError(f"Schema already exists: {schema_name}")


scaffold_my_app_name_schema = Scaffolder(
    name="scaffold-my-app-name-schema",
    input=[
        existing_module_input,
        new_entity_input,
        new_entity_column_input,
    ],
    source_path=os.path.join(os.path.dirname(__file__), "schema.template.py"),
    render_source_path=False,
    destination_path=lambda ctx: os.path.join(
        APP_DIR,
        "schema",
        f"{to_snake_case(ctx.input.entity)}.py",
    ),
    transform_content={
        "MyEntity": "{to_pascal_case(ctx.input.entity)}",
        "my_column": "{to_snake_case(ctx.input.column)}",
    },
    retries=0,
    upstream=validate_create_my_app_name_entity,
)

scaffold_my_app_name_module_entity = Scaffolder(
    name="scaffold-my-app-name-schema",
    input=[
        existing_module_input,
        new_entity_input,
        plural_entity_input,
    ],
    source_path=os.path.join(os.path.dirname(__file__), "module_template"),
    render_source_path=False,
    destination_path=lambda ctx: os.path.join(
        APP_DIR,
        "module",
        f"{to_snake_case(ctx.input.module)}",
    ),
    transform_content={
        "my_module": "{to_snake_case(ctx.input.module)}",
        "MyEntity": "{to_pascal_case(ctx.input.entity)}",
        "my_entity": "{to_snake_case(ctx.input.entity)}",
        "my_entities": "{to_snake_case(ctx.input.plural)}",
        "my-entities": "{to_kebab_case(ctx.input.plural)}",
    },
    retries=0,
    upstream=validate_create_my_app_name_entity,
)


@make_task(
    name="register-my-app-name-migration",
    input=[existing_module_input, new_entity_input],
    retries=0,
    upstream=validate_create_my_app_name_entity,
)
async def register_my_app_name_migration(ctx: AnyContext):
    migration_metadata_file_path = os.path.join(
        APP_DIR, "module", to_snake_case(ctx.input.module), "migration_metadata.py"
    )
    app_name = os.path.basename(APP_DIR)
    with open(migration_metadata_file_path, "r") as f:
        file_content = f.read()
    entity_name = to_snake_case(ctx.input.entity)
    entity_class = to_pascal_case(ctx.input.entity)
    new_file_content_list = (
        [f"from {app_name}.schema.{entity_name} import {entity_class}"]
        + file_content.strip()
        + [
            f"{entity_class}.metadata = metadata",
            f"{entity_class}.__table__.tometadata(metadata)",
            "",
        ]
    )
    with open(migration_metadata_file_path, "w") as f:
        f.write("\n".join(new_file_content_list))


@make_task(
    name="register-my-app-name-api-client",
    input=[existing_module_input, new_entity_input],
    retries=0,
    upstream=validate_create_my_app_name_entity,
)
async def register_my_app_name_api_client(ctx: AnyContext):
    api_client_file_path = os.path.join(
        APP_DIR, "module", to_snake_case(ctx.input.module), "client", "api_client.py"
    )
    with open(api_client_file_path, "r") as f:
        file_content = f.read()
    module_config_name = to_snake_case(ctx.input.module).upper()
    new_code = add_code_to_module(
        file_content,
        f"user_api_client = user_usecase.as_api_client(base_url=APP_{module_config_name}_BASE_URL)",  # noqa
    )
    new_code = add_parent_to_class(
        original_code=new_code,
        class_name="APIClient",
        parent_class_name="user_api_client",
    )
    app_name = os.path.basename(APP_DIR)
    entity_name = to_snake_case(ctx.input.entity)
    module_name = to_snake_case(ctx.input.module)
    new_file_content_list = [
        f"from {app_name}.module.{module_name}.service.{entity_name} import {entity_name}_usecase",  # noqa
        new_code.strip(),
        "",
    ]
    with open(api_client_file_path, "w") as f:
        f.write("\n".join(new_file_content_list))


@make_task(
    name="register-my-app-name-direct-client",
    input=[existing_module_input, new_entity_input],
    retries=0,
    upstream=validate_create_my_app_name_entity,
)
async def register_my_app_name_direct_client(ctx: AnyContext):
    direct_client_file_path = os.path.join(
        APP_DIR, "module", to_snake_case(ctx.input.module), "client", "direct_client.py"
    )
    with open(direct_client_file_path, "r") as f:
        file_content = f.read()
    new_code = add_code_to_module(
        file_content, "user_direct_client = user_usecase.as_direct_client()"
    )
    new_code = add_parent_to_class(
        original_code=new_code,
        class_name="DirectClient",
        parent_class_name="user_direct_client",
    )
    app_name = os.path.basename(APP_DIR)
    entity_name = to_snake_case(ctx.input.entity)
    module_name = to_snake_case(ctx.input.module)
    new_file_content_list = [
        f"from {app_name}.module.{module_name}.service.{entity_name} import {entity_name}_usecase",  # noqa
        new_code.strip(),
        "",
    ]
    with open(direct_client_file_path, "w") as f:
        f.write("\n".join(new_file_content_list))


create_my_app_name_entity = app_create_group.add_task(
    Task(
        name="create-my-app-name-entity",
        description="🏗️ Create new entity on a module",
        successor="",
    ),
    alias="entity",
)
create_my_app_name_entity << [
    scaffold_my_app_name_schema,
    scaffold_my_app_name_module_entity,
    register_my_app_name_api_client,
    register_my_app_name_direct_client,
]



from fastapp_template.common.db_engine import engine
from fastapp_template.config import APP_REPOSITORY_TYPE
from fastapp_template.module.my_module.service.my_entity.repository.my_entity_db_repository import (
    MyEntityDBRepository,
)
from fastapp_template.module.my_module.service.my_entity.repository.my_entity_repository import (
    MyEntityRepository,
)

if APP_REPOSITORY_TYPE == "db":
    my_entity_repository: MyEntityRepository = MyEntityDBRepository(engine)
else:
    my_entity_repository: MyEntityRepository = None



from fastapp_template.module.module_template.client.any_client import BaseClient
from fastapp_template.module.module_template.service.user.usecase import user_usecase


class DirectClient(BaseClient):
    pass



from zrb.attr.type import StrAttr, StrListAttr
from zrb.context.any_shared_context import AnySharedContext
from zrb.input.base_input import BaseInput
from zrb.util.attr import get_str_list_attr


class OptionInput(BaseInput):
    def __init__(
        self,
        name: str,
        description: str | None = None,
        prompt: str | None = None,
        options: StrListAttr = [],
        default_str: StrAttr = "",
        auto_render: bool = True,
        allow_empty: bool = False,
    ):
        super().__init__(
            name=name,
            description=description,
            prompt=prompt,
            default_str=default_str,
            auto_render=auto_render,
            allow_empty=allow_empty,
        )
        self._options = options

    def to_html(self, ctx: AnySharedContext) -> str:
        name = self.name
        description = self.description
        default = self._get_default_str(ctx)
        html = [f'<select name="{name}" placeholder="{description}">']
        for value in get_str_list_attr(ctx, self._options, self._auto_render):
            selected = "selected" if value == default else ""
            html.append(f'<option value="{value}" {selected}>{value}</option>')
        html.append("</select>")
        return "\n".join(html)

    def _prompt_cli_str(self, shared_ctx: AnySharedContext) -> str:
        prompt_message = self.prompt_message
        default_value = self._get_default_str(shared_ctx)
        options = get_str_list_attr(shared_ctx, self._options, self._auto_render)
        option_str = ", ".join(options)
        if default_value != "":
            prompt_message = f"{prompt_message} ({option_str}) [{default_value}]"
        value = input(f"{prompt_message}: ")
        if value.strip() != "" and value.strip() not in options:
            value = self._prompt_cli_str(shared_ctx)
        if value.strip() == "":
            value = default_value
        return value



from abc import ABC, abstractmethod

from fastapp_template.schema.my_entity import (
    MyEntity,
    MyEntityCreate,
    MyEntityResponse,
    MyEntityUpdate,
)


class MyEntityRepository(ABC):

    @abstractmethod
    async def create(self, my_entity_data: MyEntityCreate) -> MyEntityResponse:
        pass

    @abstractmethod
    async def get_by_id(self, my_entity_id: str) -> MyEntity:
        pass

    @abstractmethod
    async def get_all(self) -> list[MyEntity]:
        pass

    @abstractmethod
    async def update(
        self, my_entity_id: str, my_entity_data: MyEntityUpdate
    ) -> MyEntity:
        pass

    @abstractmethod
    async def delete(self, my_entity_id: str) -> MyEntity:
        pass

    @abstractmethod
    async def create_bulk(
        self, my_entity_data_list: list[MyEntityCreate]
    ) -> list[MyEntityResponse]:
        pass



from fastapp_template.common.base_usecase import BaseUsecase
from fastapp_template.module.my_module.service.my_entity.repository.factory import (
    my_entity_repository,
)
from fastapp_template.module.my_module.service.my_entity.repository.repository import (
    MyEntityRepository,
)
from fastapp_template.schema.my_entity import (
    MyEntityCreate,
    MyEntityResponse,
    MyEntityUpdate,
)


class MyEntityUsecase(BaseUsecase):

    def __init__(self, my_entity_repository: MyEntityRepository):
        super().__init__()
        self.my_entity_repository = my_entity_repository

    @BaseUsecase.route(
        "/api/v1/my-entities/{my_entity_id}",
        methods=["get"],
        response_model=MyEntityResponse,
    )
    async def get_my_entity_by_id(self, my_entity_id: str) -> MyEntityResponse:
        return await self.my_entity_repository.get_by_id(my_entity_id)

    @BaseUsecase.route(
        "/api/v1/my-entities", methods=["get"], response_model=list[MyEntityResponse]
    )
    async def get_all_my_entities(self) -> list[MyEntityResponse]:
        return await self.my_entity_repository.get_all()

    @BaseUsecase.route(
        "/api/v1/my-entities",
        methods=["post"],
        response_model=MyEntityResponse | list[MyEntityResponse],
    )
    async def create_my_entity(
        self, data: MyEntityCreate | list[MyEntityCreate]
    ) -> MyEntityResponse | list[MyEntityResponse]:
        if isinstance(data, MyEntityCreate):
            return await self.my_entity_repository.create(data)
        return await self.my_entity_repository.create_bulk(data)

    @BaseUsecase.route(
        "/api/v1/my-entities/{my_entity_id}",
        methods=["put"],
        response_model=MyEntityResponse,
    )
    async def update_my_entity(
        self, my_entity_id: str, data: MyEntityUpdate
    ) -> MyEntityResponse:
        return await self.my_entity_repository.update(my_entity_id, data)

    @BaseUsecase.route(
        "/api/v1/my-entities/{my_entity_id}",
        methods=["delete"],
        response_model=MyEntityResponse,
    )
    async def delete_my_entity(self, my_entity_id: str) -> MyEntityResponse:
        return await self.my_entity_repository.delete(my_entity_id)


my_entity_usecase = MyEntityUsecase(my_entity_repository=my_entity_repository)



import os

from fastapp_template._zrb.config import APP_DIR
from fastapp_template._zrb.group import app_create_group
from fastapp_template._zrb.helper import get_existing_module_names
from fastapp_template._zrb.input import new_module_input

from zrb import AnyContext, Scaffolder, Task, make_task
from zrb.util.string.conversion import to_kebab_case, to_pascal_case, to_snake_case


@make_task(
    name="validate-create-my-app-name-module",
    input=new_module_input,
    retries=0,
)
async def validate_create_my_app_name_module(ctx: AnyContext):
    if ctx.input.module in get_existing_module_names():
        raise ValueError(f"Module already exists: {ctx.input.module}")


scaffold_my_app_name_module = Scaffolder(
    name="scaffold-my-app-name-module",
    input=new_module_input,
    source_path=os.path.join(os.path.dirname(__file__), "module_template"),
    render_source_path=False,
    destination_path=lambda ctx: os.path.join(
        APP_DIR,
        "module",
        to_snake_case(ctx.input.module),
    ),
    transform_content={
        "module_template": "{to_snake_case(ctx.input.module)}",
        "MY_MODULE_NAME": "{to_snake_case(ctx.input.module).upper()}",
    },
    retries=0,
    upstream=validate_create_my_app_name_module,
)


@make_task(
    name="register-my-app-name-module-config",
    input=new_module_input,
    upstream=validate_create_my_app_name_module,
    retries=0,
)
async def register_my_app_name_module_config(ctx: AnyContext):
    """Registering module to config.py"""
    existing_module_names = get_existing_module_names()
    module_port = 3000 + len(
        [
            module_name
            for module_name in existing_module_names
            if module_name != to_snake_case(ctx.input.module)
        ]
    )
    module_base_url = f"http://localhost:{module_port}"
    config_file_name = os.path.join(APP_DIR, "config.py")
    upper_module_name = to_snake_case(ctx.input.module).upper()
    config_name = f"APP_{upper_module_name}_BASE_URL"
    env_name = f"MY_APP_NAME_{upper_module_name}_BASE_URL"
    # TODO: check before write
    with open(config_file_name, "a") as f:
        f.write(f'{config_name} = os.getenv("{env_name}", "{module_base_url}")\n')


@make_task(
    name="register-my-app-name-module",
    input=new_module_input,
    upstream=validate_create_my_app_name_module,
    retries=0,
)
async def register_my_app_name_module(ctx: AnyContext):
    """Registering module to application's main.py"""
    app_main_file_name = os.path.join(APP_DIR, "main.py")
    module_name = to_snake_case(ctx.input.module)
    import_code = f"from fastapp_template.module.{module_name} import route as {module_name}_route"  # noqa
    assert_code = f"assert {module_name}_route"
    with open(app_main_file_name, "r") as f:
        code = f.read()
    new_code = "\n".join([import_code, code.strip(), assert_code, ""])
    # TODO: check before write
    with open(app_main_file_name, "w") as f:
        f.write(new_code)


@make_task(
    name="register-my-app-name-module-runner",
    input=new_module_input,
    upstream=validate_create_my_app_name_module,
    retries=0,
)
async def register_my_app_name_module_runner(ctx: AnyContext):
    """Registering module to _zrb's main.py"""
    task_main_file_name = os.path.join(APP_DIR, "_zrb", "main.py")
    existing_module_names = get_existing_module_names()
    module_port = 3001 + len(
        [
            module_name
            for module_name in existing_module_names
            if module_name != to_snake_case(ctx.input.module)
        ]
    )
    module_snake_name = to_snake_case(ctx.input.module)
    module_kebab_name = to_kebab_case(ctx.input.module)
    module_pascal_name = to_pascal_case(ctx.input.module)
    with open(os.path.join(os.path.dirname(__file__), "run_module.template.py")) as f:
        module_runner_code = (
            f.read()
            .replace("my_module", module_snake_name)
            .replace("my-module", module_kebab_name)
            .replace("My Module", module_pascal_name)
            .replace("3000", f"{module_port}")
        )
    with open(task_main_file_name, "r") as f:
        code = f.read()
    new_code = "\n".join([code.strip(), "", module_runner_code, ""])
    # TODO: check before write
    with open(task_main_file_name, "w") as f:
        f.write(new_code)


create_my_app_name_module = app_create_group.add_task(
    Task(
        name="create-my-app-name-module",
        description="🧩 Create new module on My App Name",
        retries=0,
    ),
    alias="module",
)
create_my_app_name_module << [
    scaffold_my_app_name_module,
    register_my_app_name_module,
    register_my_app_name_module_config,
    register_my_app_name_module_runner,
]



from fastapp_template.common.base_db_repository import BaseDBRepository
from fastapp_template.common.error import NotFoundError
from fastapp_template.module.my_module.service.my_entity.repository.my_entity_repository import (
    MyEntityRepository,
)
from fastapp_template.schema.my_entity import (
    MyEntity,
    MyEntityCreate,
    MyEntityResponse,
    MyEntityUpdate,
)
from passlib.context import CryptContext
from sqlalchemy.ext.asyncio import AsyncSession
from sqlmodel import Session, select

# Password hashing context
pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


def hash_password(password: str) -> str:
    return pwd_context.hash(password)


class MyEntityDBRepository(
    BaseDBRepository[MyEntity, MyEntityResponse, MyEntityCreate, MyEntityUpdate],
    MyEntityRepository,
):
    db_model = MyEntity
    response_model = MyEntityResponse
    create_model = MyEntityCreate
    update_model = MyEntityUpdate
    entity_name = "my_entity"
    column_preprocessors = {}



import asyncio
import os
import re
import sys
from collections.abc import Callable

from zrb.cmd.cmd_result import CmdResult


def check_unrecommended_commands(cmd_script: str) -> dict[str, str]:
    banned_commands = {
        "<(": "Process substitution isn't POSIX compliant and causes trouble",
        "column": "Command isn't included in Ubuntu packages and is not POSIX compliant",
        "echo": "echo isn't consistent across OS; use printf instead",
        "eval": "Avoid eval as it can accidentally execute arbitrary strings",
        "realpath": "Not available by default on OSX",
        "source": "Not POSIX compliant; use '.' instead",
        " test": "Use '[' instead for consistency",
        "which": "Command in not POSIX compliant, use command -v",
    }
    banned_commands_regex = {
        r"grep.* -y": "grep -y does not work on Alpine; use grep -i",
        r"grep.* -P": "grep -P is not valid on OSX",
        r"grep[^|]+--\w{2,}": "grep long commands do not work on Alpine",
        r'readlink.+-.*f.+["$]': "readlink -f behaves differently on OSX",
        r"sort.*-V": "sort -V is not supported everywhere",
        r"sort.*--sort-versions": "sort --sort-version is not supported everywhere",
        r"\bls ": "Avoid using ls; use shell globs or find instead",
    }
    violations = {}
    # Check banned commands
    for cmd, reason in banned_commands.items():
        if cmd in cmd_script:
            violations[cmd] = reason
    # Check banned regex patterns
    for pattern, reason in banned_commands_regex.items():
        if re.search(pattern, cmd_script):
            violations[pattern] = reason
    return violations


async def run_command(
    cmd: list[str],
    cwd: str | None = None,
    env_map: dict[str, str] | None = None,
    log_method: Callable[..., None] = print,
    max_output_line: int = 1000,
    max_error_line: int = 1000,
) -> tuple[CmdResult, int]:

    async def __read_stream(
        stream, log_method: Callable[..., None], max_lines: int
    ) -> str:
        lines = []
        while True:
            line = await stream.readline()
            if not line:
                break
            line = line.decode("utf-8").rstrip()
            lines.append(line)
            if len(lines) > max_lines:
                lines.pop(0)  # Keep only the last max_lines
            log_method(line)
        return "\n".join(lines)

    try:
        if cwd is None:
            cwd = os.getcwd()
        if env_map is None:
            env_map = os.environ
        cmd_process = await asyncio.create_subprocess_exec(
            *cmd,
            cwd=cwd,
            stdin=sys.stdin if sys.stdin.isatty() else None,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=env_map,
            bufsize=0,
        )
        stdout_task = asyncio.create_task(
            __read_stream(cmd_process.stdout, log_method, max_output_line)
        )
        stderr_task = asyncio.create_task(
            __read_stream(cmd_process.stderr, log_method, max_error_line)
        )
        # Wait for process to complete and gather stdout/stderr
        return_code = await cmd_process.wait()
        stdout = await stdout_task
        stderr = await stderr_task
        return CmdResult(stdout, stderr), return_code
    finally:
        if cmd_process is not None and cmd_process.returncode is None:
            cmd_process.terminate()



from fastapp_template._zrb.group import app_create_group

from zrb import Task

create_my_app_name_column = app_create_group.add_task(
    Task(
        name="create-my-app-name-column",
        description="📊 Create new column on an entity",
    ),
    alias="column",
)



def add_code_to_module(source_code: str, new_code: str) -> str:
    lines = source_code.splitlines()
    last_import_index = -1
    for i, line in enumerate(lines):
        stripped_line = line.strip()
        if stripped_line.startswith("import") or stripped_line.startswith("from"):
            last_import_index = i
        elif stripped_line and not stripped_line.startswith("#"):
            break
    if last_import_index != -1:
        lines.insert(last_import_index + 1, new_code)
    return "\n".join(lines)
